import sys
import click
from snake.game import Game, GameOptions


@click.command()
@click.option("--engine",
              default=GameOptions.MODE_ASCIIMATICS,
              type=click.Choice([GameOptions.MODE_BASIC, GameOptions.MODE_ASCIIMATICS]),
              help="The game engine: determine how the game looks and behave. "
                   "Choose `{}` for the best game experience, `{}` for better portability."
                   .format(GameOptions.MODE_ASCIIMATICS, GameOptions.MODE_BASIC))
@click.option("--wall-mode",
              default="pacman",
              type=click.Choice([GameOptions.WALL_PACMAN, GameOptions.WALL_WALL]),
              help="The walls mode. If `{}` mode is chose the snake will pass through walls, otherwise if "
                   "`{}` is chosen the snake will die hitting the wall."
                   .format(GameOptions.WALL_PACMAN, GameOptions.WALL_WALL))
@click.option("--rows-no",
              default=10,
              type=click.IntRange(2, 1000),
              help="The board width. (Please note that in terminal vertical chars size is larger than the horizontal "
                   "one, this may seem to make the snake go faster. Try to keep `rows-no` half of `cols-no`).")
@click.option("--cols-no",
              default=30,
              type=click.IntRange(2, 1000),
              help="The board height. (Please note that in terminal vertical chars size is larger than the horizontal "
                   "one, this may seem to make the snake go faster. Try to keep `cols-no` twice bigger than `rows-no`).")
@click.option("--speed",
              default=0,
              type=click.INT,
              help="The snake initial speed. An integer from 0 to N, where 0 is the minimum speed and N the maximum ("
                   "technically there is no speed limit)")
@click.option("--initial-length",
              default=3,
              type=click.INT,
              help="The snake initial length. By default the snake is 3 blocks longer.")
@click.option("--eat-multiplier",
              default=1,
              type=click.INT,
              help="An integer representing the eat multiplier. The snake enlarges by `eat-multiplier` every time "
                   "that eats. By default it enlarges by one.")
@click.option("--food-no",
              default=3,
              type=click.INT,
              help="Food number on board.")
def main(engine, wall_mode, rows_no, cols_no, speed, food_no, eat_multiplier, initial_length):
    """
    Welcome to snake on terminal!

    Luca Parolari  <luca.parolari23@gmail.com>

    Snake on terminal is a simple implementation of the famous arcade game, on terminal obviously.
    The game is available with several interchangeable engines that modify how the game looks like and behaves.
    See the `--engine` options for more information.

    Drive the snake with 'w', 'a', 's', 'd'. Use 'q' to quit.

    Enjoy!
    """

    options = GameOptions(mode=engine,
                          wall_mode=wall_mode,
                          speed=speed,
                          rows=rows_no,
                          cols=cols_no,
                          foods=food_no,
                          eat_multiplier=eat_multiplier,
                          initial_length=initial_length)

    game = Game(options)

    try:
        game.play()
    except Exception as e:
        print(e)

    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import numpy

from snake.broadcaster import broadcaster
from snake.keyb import keyboardBuffer
from snake.util import wasd_to_direction


class Entity(object):
    """
    This class conceptually represent a single or composite entity on matrix.
    Provide a method used to inform the matrix with the new values on board.
    """
    def to_list(self):
        raise NotImplementedError()


class GameOverException(Exception):
    pass


class Board(object):

    SNAKE_HEAD_REPR = 1.0
    SNAKE_TAIL_REPR = 2.0
    EMPTY_REPR = 0.0
    FOOD_REPR = 3.0

    _rows_no = 0
    _cols_no = 0
    _food_no = 0

    _snake = None
    _food = None

    _matrix = numpy.empty(())

    def __init__(self, rows_no, cols_no, food_no, snake_entity, food_entity):
        self._rows_no = rows_no
        self._cols_no = cols_no
        self._food_no = food_no
        self._snake = snake_entity
        self._food = food_entity

        # initialization
        self._update_status()  # draw the snake on matrix
        self._food.more(food_no, self._empty_points())  # add flowers
        self._update_status()  # draw also flowers on matrix

        broadcaster.listen("snake_eat", self._snake_eat)
        broadcaster.listen("snake_moved", self._snake_moved)

    def step(self):
        self._update_input()
        self._update_status()

    def matrix(self):
        """
        Build the game matrix with internal representation.
        :return:
        """
        return self._matrix.copy()

    def _update_status(self):
        """
        Update the status matrix by side effect and returns it.
        :return: A new status matrix.
        """
        self._matrix = numpy.full((self._rows_no, self._cols_no), Board.EMPTY_REPR)

        s = self._snake.to_list()

        for i in range(len(s)):
            si, sj = s[i]
            if i == 0:
                self._matrix[si][sj] = self.SNAKE_HEAD_REPR
            else:
                self._matrix[si][sj] = self.SNAKE_TAIL_REPR

        f = self._food.to_list()

        for i in range(len(f)):
            fi, fj = f[i]
            self._matrix[fi][fj] = self.FOOD_REPR

    def _update_input(self):
        """
        Keep the game alive updating things based on inputs.
        :return:
        """
        # keyboard buffer is necessary to keep the pressed keys history between discrete steps of the games.
        key = keyboardBuffer.next()

        if key == "q":
            raise GameOverException

        direction = wasd_to_direction(key)

        self._snake.move(direction)

    def _snake_moved(self, args):
        """
        Callback for snake moves.
        :param args: A tuple containing in the first element the parameter head, i.e. the head of the snake.
        """
        head = args[0]

        if head in self._snake.to_list()[1:]:
            raise GameOverException()
        if head in self._food.to_list():
            self._snake.eat()

    def _snake_eat(self, args):
        """
        Callback for snake eat.
        :param args: A tuple containing in the first element the parameter head, i.e. the head of the snake.
        :return:
        """
        head = args[0]

        self._food.eat(head)
        self._food.one(self._empty_points())

    def _empty_points(self):
        """
        :return: A list of empty points from the board.
        """
        point = []
        for i in range(self._rows_no):
            for j in range(self._cols_no):
                if self._matrix[i][j] == self.EMPTY_REPR:
                    point.append((i, j))
        return point



import random

from snake.board import Entity


class Food(Entity):

    _flowers = []

    def one(self, available):
        self._flowers.append(available[random.randrange(0, len(available))])

    def more(self, n, available):
        for i in range(n):
            self.one(available)

    def eat(self, food):
        self._flowers.remove(food)

    def to_list(self):
        return self._flowers.copy()



import time

import numpy
from asciimatics.screen import Screen

from snake.board import Board
from snake.broadcaster import broadcaster


class Engine(object):
    _speed = 0

    def __init__(self, board):
        self._board = board
        broadcaster.listen("snake_eat", self._inc_speed)

    def start(self):
        raise NotImplementedError()

    def draw(self, *args, **kwargs):
        raise NotImplementedError()

    @property
    def speed(self):
        """
        :return: The computed game update speed.
        """
        return numpy.exp(-self._speed / 10) / 2

    @speed.setter
    def speed(self, v):
        self._speed = v

    def _inc_speed(self, _args):
        self._speed += 1


class BasicEngine(Engine):

    def start(self):
        while True:
            self._board.step()
            self.draw(matrix=self._board.matrix())

            time.sleep(self.speed)

    def draw(self, *args, **kwargs):
        matrix = kwargs['matrix']

        cast = {Board.EMPTY_REPR: " ", Board.SNAKE_HEAD_REPR: "@", Board.SNAKE_TAIL_REPR: "#", Board.FOOD_REPR: "+"}

        (r, c) = matrix.shape
        print("*" * (c + 2))
        for i in range(r):
            print("*", end='')
            for j in range(c):
                print(
                    cast[matrix[i][j]],
                    end='')
            print("*")
        print("*" * (c + 2))


class AsciimaticsEngine(Engine):

    def start(self):
        Screen.wrapper(self.draw)

    def draw(self, *args, **kwargs):
        while True:
            self._board.step()
            matrix = self._board.matrix()
            screen = args[0]

            cast = {Board.EMPTY_REPR: " ", Board.SNAKE_HEAD_REPR: "@", Board.SNAKE_TAIL_REPR: "#", Board.FOOD_REPR: "+"}

            (r, c) = matrix.shape

            screen.print_at("*" * (c + 2), 0, 0)
            for i in range(r):
                line = "*"
                for j in range(c):
                    line += cast[matrix[i][j]]
                line += "*"
                screen.print_at(line, 0, i+1)
            screen.print_at("*" * (c + 2), 0, r + 1)

            screen.refresh()

            time.sleep(self.speed)



from snake.board import Entity, GameOverException
from snake.broadcaster import broadcaster


class NormalizationStrategy(object):

    def _normalize(self, coord):
        raise NotImplementedError()

    def get(self):
        return self._normalize


class PacmanNormalizationStrategy(NormalizationStrategy):

    _boundary = None

    def __init__(self, boundary):
        self._boundary = boundary

    def _normalize(self, coord):
        """
        :param coord: The (i, j) coords tuple to normalize.
        :return: A tuple (i, j) normalized with a normalization method.
        """
        rows_no, cols_no = self._boundary
        i, j = coord

        i = (i + rows_no) % rows_no
        j = (j + cols_no) % cols_no

        return i, j


class WallNormalizationStrategy(NormalizationStrategy):
    _boundary = None

    def __init__(self, boundary):
        self._boundary = boundary

    def _normalize(self, coord):
        """
        :param coord: The (i, j) coords tuple to normalize.
        :return: A tuple (i, j) normalized with a normalization method.
        """
        wi, wj = self._boundary
        i, j = coord

        if i < 0 or i > wi or j < 0 or j > wj:
            raise GameOverException()

        return i, j


class Snake(Entity):

    NORM_MODE_PACMAN = "pacman"
    NORM_MODE_WALL = "wall"

    # head, t_1, .., t_n
    _edges = []
    _coords = 0, 0

    _eat_multiplier = 1

    _normalize = None

    def __init__(self, initial_length, eat_multiplier):
        self._edges = ["right" for _i in range(initial_length)]
        self._coords = (0, initial_length)
        self._eat_multiplier = eat_multiplier

    def move(self, direction):
        """
        Move the snake by the given direction, if possible. Check the input directions.
        :param direction: The new direction of the snake.
        """
        # input validation
        if direction is None or direction not in ("up", "down", "left", "right"):
            direction = self._last_move()

        # direction validation
        if (direction == "up" and self._last_move() == "down"
                or direction == "down" and self._last_move() == "up"
                or direction == "left" and self._last_move() == "right"
                or direction == "right" and self._last_move() == "left"):
            direction = self._last_move()

        # increment head coords
        i, j = self._coords
        di, dj = self._inc(direction)
        self._coords = self._normalize((i + di, j + dj))

        # add edge
        self._edges.insert(0, direction)
        self._edges.pop(-1)

        # moved event
        broadcaster.event("snake_moved", self._coords)

    def eat(self):
        """
        Make the snake longer by one.
        """
        last = self._edges[-1]

        for i in range(self._eat_multiplier):
            self._edges.append(last)

        # eat event
        d, hi, hj = self._head()
        broadcaster.event("snake_eat", (hi, hj))

    def set_normalization_strategy(self, normalization_strategy):
        self._normalize = normalization_strategy

    def _head(self):
        """
        :return: A tuple with (direction, i, j) where i, j are the head coords
        """
        i, j = self._coords
        return self._edges[0], i, j

    def _tail(self):
        """
        :return: A list of directions.
        """
        tail = self._edges[1:]
        return tail

    def _last_move(self):
        """
        :return: The last direction.
        """
        head, i, j = self._head()
        return head

    def to_list(self):
        """
        :return: A list of edges with coordinates where the first is the snake head and the rest the snake tail.
        """
        ll = []

        head, ti, tj = self._head()
        tail = self._tail()

        # head
        ll.append((ti, tj))

        di, dj = self._dec(head)
        ti, tj = self._normalize((ti + di, tj + dj))

        # tail
        for s in tail:
            ll.append((ti, tj))

            di, dj = self._dec(s)
            ti, tj = self._normalize((ti + di, tj + dj))

        return ll

    @staticmethod
    def _inc(direction):
        """
        :param direction: The direction, one in [up, down, left, right].
        :return: The tuple (delta i, delta j) based on direction with increment by one.
        """
        di, dj = 0, 0
        if direction == "up":
            di = -1
        if direction == "down":
            di = +1
        if direction == "left":
            dj = -1
        if direction == "right":
            dj = +1
        return di, dj

    @staticmethod
    def _dec(direction):
        """
        :param direction: The direction, one in [up, down, left, right].
        :return: The tuple (delta i, delta j) based on direction with decrement by one.
        """
        i, j = Snake._inc(direction)
        return -i, -j



import click

from snake.board import Board, GameOverException
from snake.broadcaster import broadcaster
from snake.engine import AsciimaticsEngine, BasicEngine
from snake.food import Food
from snake.keyb import keyboardManager, keyboardBuffer
from snake.snaky import Snake, PacmanNormalizationStrategy, WallNormalizationStrategy


class GameOptions(object):
    MODE_BASIC = 'basic'
    MODE_ASCIIMATICS = 'asciimatics'
    WALL_PACMAN = 'pacman'
    WALL_WALL = 'wall'

    _mode = None
    _wall_mode = None
    _rows = None
    _cols = None
    _speed = None
    _foods = None
    _eat_multiplier = None
    _initial_length = None

    def __init__(self,
                 mode=MODE_BASIC,
                 wall_mode=WALL_PACMAN,
                 rows=10, cols=10,
                 speed=0,
                 foods=3,
                 eat_multiplier=1,
                 initial_length=3):
        """
        Initialize game options. If some value is semantically incorrect it will be replaced with defaults.
        """
        if mode not in [self.MODE_BASIC, self.MODE_ASCIIMATICS]:
            mode = self.MODE_BASIC
        if rows < 10:
            rows = 10
        if cols < 10:
            cols = 10
        if speed < 0:
            speed = 0
        if foods < 0:
            foods = 0
        if eat_multiplier < 0:
            eat_multiplier = 0
        if initial_length < 1:
            initial_length = 1

        self._mode = mode
        self._wall_mode = wall_mode
        self._rows = rows
        self._cols = cols
        self._speed = speed
        self._foods = foods
        self._eat_multiplier = eat_multiplier
        self._initial_length = initial_length

    @property
    def mode(self):
        return self._mode

    @property
    def wall_mode(self):
        return self._wall_mode

    @property
    def rows(self):
        return self._rows

    @property
    def cols(self):
        return self._cols

    @property
    def speed(self):
        return self._speed

    @property
    def foods(self):
        return self._foods

    @property
    def eat_multiplier(self):
        return self._eat_multiplier

    @property
    def initial_length(self):
        return self._initial_length


class Game(object):

    _score = 0

    def __init__(self, options):
        # normalization strategies.
        normalization_strategy = PacmanNormalizationStrategy((options.rows, options.cols))
        if options.wall_mode == GameOptions.WALL_WALL:
            normalization_strategy = WallNormalizationStrategy((options.rows, options.cols))

        # game objects
        snake = Snake(initial_length=options.initial_length, eat_multiplier=options.eat_multiplier)
        snake.set_normalization_strategy(normalization_strategy.get())
        food = Food()
        board = Board(options.rows, options.cols, options.foods, snake, food)

        # engine
        engine = BasicEngine(board=board)
        if options.mode == GameOptions.MODE_ASCIIMATICS:
            engine = AsciimaticsEngine(board=board)

        self._engine = engine
        self._engine.speed = options.speed

        # callbacks
        keyboardManager.register("w", lambda: keyboardBuffer.bufferize("w"))
        keyboardManager.register("a", lambda: keyboardBuffer.bufferize("a"))
        keyboardManager.register("s", lambda: keyboardBuffer.bufferize("s"))
        keyboardManager.register("d", lambda: keyboardBuffer.bufferize("d"))
        keyboardManager.register("q", lambda: keyboardBuffer.bufferize("q"))

        broadcaster.listen("snake_eat", self._inc_score)

    def play(self):
        try:
            self._engine.start()
        except GameOverException as _e:
            self.game_over()

    def game_over(self):
        click.echo("###########################################")
        click.echo("                GAME OVER                 ")
        click.echo("                                          ")
        click.echo(" Your score is {},".format(self._score))
        click.echo(" {}".format(self._motivational_phrase(self._score)))
        click.echo("###########################################")
        # click.echo("")
        # click.echo("  ---")
        # click.echo("  Thanks for playing SNAKE ON TERMINAL")
        # click.echo("  Luca Parolari <luca.parolari23@gmail.com>")
        # click.echo("")
        # click.echo("  If you linked this game and/or the code behind this please leave a star")
        # click.echo("  to the github repository. If you find a bug let me know through the ")
        # click.echo("  issue tab on github.")
        # click.echo("  Link: https://github.com/lparolari/snake")

    @staticmethod
    def _motivational_phrase(score):
        if score in range(0, 8):
            return "you need to practise more ;)"
        if score in range(9, 16):
            return "wow, you are learning!"
        if score in range(17, 30):
            return "good game! :D"
        if score in range(31, 50):
            return "you are a pro gamer!"
        if score in range(51, 100):
            return "are you cheating? o.O"
        if score > 100:
            return "God mode on!"

    def _inc_score(self, _args):
        self._score += 1



from pynput.keyboard import Listener, KeyCode


class KeyboardManager(object):
    _callbacks = {}
    _listener = None

    def __init__(self):
        pass
        self._listener = Listener(on_press=self._on_press)
        self._listener.start()

    def _on_press(self, key):
        if isinstance(key, KeyCode):
            key = key.char
        self._execute(key)

    def _execute(self, key):
        callback = self._callbacks.get(key)
        if callback is None:
            return
        try:
            callback()
        except:
            # catch-all for keyboard reactions
            pass

    def register(self, key, callback):
        """
        Register the callback 'callback' on the key 'key'.
        """
        if isinstance(key, list):
            map(lambda k: self._callbacks.update({k: callback}), key)
        else:
            self._callbacks.update({key: callback})


class KeyboardBuffer:
    _buffer = []

    def bufferize(self, key):
        """
        Bufferize the key 'key'.
        """
        self._buffer.insert(0, key)

    def next(self):
        """
        :return: The next buffered key.
        """
        if len(self._buffer) == 0:
            return None
        return self._buffer.pop()


keyboardManager = KeyboardManager()
keyboardBuffer = KeyboardBuffer()



def wasd_to_direction(wasd):
    direction = None
    if wasd == "w":
        direction = "up"
    if wasd == "s":
        direction = "down"
    if wasd == "a":
        direction = "left"
    if wasd == "d":
        direction = "right"
    return direction



class Broadcaster(object):

    _events = {}

    def event(self, e, *args):
        # TODO: catch exception, but... they are needed
        # The game terminates its execution with an exception and the exception is fired in the board.py file
        # during a callback invoked within this even notifier.
        # A solution could be: change the listen method in order to accept a callbackException, called in case
        # of exception with the proper object and catch all unintended exceptions.

        cs = self._events.get(e) if self._events.get(e) is not None else []
        for c in cs:
            c(args)

    def listen(self, to, callback):
        if self._events.get(to) is None:
            d = {to: [callback]}
        else:
            callbacks = self._events.get(to).copy()
            callbacks.append(callback)
            d = {to: callbacks}
        self._events.update(d)


broadcaster = Broadcaster()
